import re
import inspect

"""
Authors: www.iExpertos.com
   - Carlos Ble (www.carlosble.com) 

License: Apache 2 (http://www.apache.org/licenses/LICENSE-2.0.html)

Project home: http://www.pydoubles.org
"""

ANY_ARG = "<___ANY___ARG___>"

class UnexpectedBehavior(Exception):
    pass

class WrongApiUsage(Exception):
    pass

class ArgsDontMatch(Exception):
    pass

class ApiMismatch(Exception):
    pass


class _MethodHandler_():
    """
    Internal framework class. When user access a method or
    any other property in the test double, the framework
    return an instance of this class, which is callable.
    So calling this instance with parenthesis triggers
    the on_method_call event in the test double.
    """    
    
    def __init__(self, double, attr):
        self.double = double
        self.attr_name = attr
        
    def __call__(self, *args, **kwargs):
        self.double.invoked_method_name = self.attr_name
        return self.double._on_method_call(self, *args, 
                                                  **kwargs)
    
    
class _Introspector_():
    """
    Internal framework class. Specialized in introspection (reflection)
    """    
    
    def method_name(self, interceptor):
        return interceptor.attr_name
    
    def double_instance_from_method(self, interceptor):
        try:
            return interceptor.double
        except AttributeError,e:
            raise WrongApiUsage("Make sure you call this framework method passing in a method from a spy or a mock object")
    
    def original_method_signature(self, original_method):
        arg_spec = inspect.getargspec(original_method)
        kargs_spec_len = 0
        args_plus_kwargs_spec_len = len(arg_spec.args) -1
        if arg_spec.defaults is not None:
           kargs_spec_len = len(arg_spec.defaults)
        return (args_plus_kwargs_spec_len, kargs_spec_len, arg_spec)
    
    def are_synonymous(self, instance, method_name1, method_name2):
        return getattr(instance, 
                       method_name1) == getattr(instance, 
                                                method_name2)

        
class _CallsRegistry_():
    """
    Internal framework class. It is used to store what calls
    are made in a test double
    """    
    
    def __init__(self):
        self.calls_args_per_method = {}
        self.args_matcher = _ArgsMatcher_()
        
    def _args(self, args):
        return "args=" + str(args)
    
    def _kwargs(self, kwargs):
        return "kwargs=" + str(kwargs)
    
    def register_call(self, method_name, *args, **kwargs):
        if not self.calls_args_per_method.has_key(method_name):
            self.calls_args_per_method[method_name] = []
        self.calls_args_per_method[method_name].append(self._args(args))
        self.calls_args_per_method[method_name].append(self._kwargs(kwargs))
         
    def was_registered(self, method_name):
        return method_name in self.calls_args_per_method.keys()
  
    def _readable_kwargs(self, kwargs_str):
        if kwargs_str == "{}":
            return "No keyword args where passed in"
        return kwargs_str
      
    def _format_err_msg(self, method_name, args_str, kwargs_str):
        return "RECORDED calls were: << %s >>, \n EXPECTED call is << args = %s, keyword args = %s >>" % (
                str(self.calls_args_per_method[method_name]), 
                args_str, self._readable_kwargs(kwargs_str))

    def match_call(self, method_name, *args, **kwargs):
        if not self.args_matcher.do_call_args_match(
                  args, self.calls_args_per_method[method_name]) or \
           self._kwargs(kwargs) not in self.calls_args_per_method[method_name]:
                raise ArgsDontMatch(self._format_err_msg(
                         method_name, str(args), str(kwargs)))
                
    def show_registered_calls(self):
        return str(self.calls_args_per_method)


        
class _Combinator_():
    """
    Internal framework class. 
    Given [1,2] return [1,2],[ANY,2],[1,ANY],[ANY,ANY]:
    all combinations of ANY and explicit args 
    """    
    
    def all_args_combinations(self, args):
        if len(args) == 0:
            return ((),)
        solutions = [args]
        solutions.extend(self._replace_non_any_with_any(args))
        return solutions

    def _replace_non_any_with_any(self, args):
        solutions = []
        for i in range(0, len(args)):
            combination = list(args)
            if combination[i] != ANY_ARG:
                combination[i] = ANY_ARG
                solutions.append(tuple(combination))
                solutions.extend(self._replace_non_any_with_any(combination))   
        return solutions

        
        
class _ArgsAdapter_():
    """
    Internal framework class. 
    """    
    def transform(self, args):
        args = str(args)
        args = re.sub(",\s", ",", args)
        args = re.sub("\s,", ",", args)
        args = args.replace("(", "")
        args = args.replace("u'", "")
        args = args.replace("'", "")
        args = args.replace("\"", "")
        args = args.replace(")", "")
        return args.split(',')
            
    def clean_call_args(self, call_args_list):
        result = []
        for args_or_kwargs in call_args_list:
            if re.search("^args=", args_or_kwargs):
                result.append(args_or_kwargs.split('args=')[-1])
        return result



class _ArgsMatcher_():
    """
    Internal framework class. 
    """    
    
    def __init__(self):
        self.combinator = _Combinator_()
        self.adapter = _ArgsAdapter_()
                    
    def method_key(self, method_name, args, kwargs):
        return method_name + str(args) + str(kwargs)
    
    def is_method_match(self, methods_dict, method_name,
                        args, kwargs):
        return self.__method_match(
                    methods_dict, method_name, args, kwargs)[0]

    def __method_match(self, methods_dict, method_name,
                          args, kwargs):
        combinations = self.combinator.all_args_combinations(args)
        for i in range(0, len(combinations)):
            key = self.method_key(method_name, combinations[i], kwargs)
            if methods_dict.has_key(key):
               return (True, methods_dict[key])
        return (False, None)
                
    def _every_arg_matches_or_expected_any_arg(self, args, expected_args):
        for i in range(0, len(args)):
            if args[i] != expected_args[i] and \
               expected_args[i] != ANY_ARG:
               return False
        return True
    
    def _do_args_tuple_match(self, args, expected_args):
        eargs = self.adapter.transform(expected_args)
        aargs = self.adapter.transform(args)
        if len(aargs) != len(eargs):
            return False
        return self._every_arg_matches_or_expected_any_arg(aargs, eargs)
    
    def do_call_args_match(self, expected_args, calls_args_list):
        for args in self.adapter.clean_call_args(calls_args_list):
            if self._do_args_tuple_match(args, expected_args):
                return True
        return False
    
    def matching_output(self, methods_dict, method_name,
                        args, kwargs):
        return self.__method_match(
                    methods_dict, method_name, args, kwargs)[1]

        
              
class _StubsRepository_():
    """
    Internal framework class. It is used to store what values
    should return in case of stub method calls in the test doubles objects
    """
    
    UNDEFINED = "undefined_____"
    
    def __init__(self):
        self.stubs_return_values = {}
        self._clear_stub_definition()
        self.args_matcher = _ArgsMatcher_()

    def _clear_stub_definition(self):
        self.input_args = self.UNDEFINED
        self.input_kwargs = ""
        self.last_stubbed_method = None
            
    def will_stub_any_input(self):
        return self.last_stubbed_method is not None and \
               self.input_args == self.UNDEFINED
               
    def is_stub_for_this_input(self, method_name, args, kwargs):
        return self.args_matcher.is_method_match(
                    self.stubs_return_values, method_name, args, kwargs) 
                                
    def _is_stub_for_any_input(self, method_name):
        return self.stubs_return_values.has_key(method_name)
        
    def is_matching_stubbed_method(self, method_name, args, kwargs):
        return self.is_stub_for_this_input(
                    method_name, args, kwargs) or \
               self._is_stub_for_any_input(method_name)
               
    def _get_output_value(self, method_name, args, kwargs):
        if self._is_stub_for_any_input(method_name):
           return self.stubs_return_values[method_name]
        else:
           return self.args_matcher.matching_output(
                    self.stubs_return_values, method_name, args, kwargs)
        
    def return_value_given_input(self, method_name, args, kwargs):
        ret_val = self._get_output_value(method_name, args, kwargs)
        if callable(ret_val):
           ret_val()
        return ret_val
                      
    def create_stub(self, method_name):
        self._clear_stub_definition()
        self.last_stubbed_method = method_name
        
    def set_input_for_last_stubbed_method(self, args, kwargs):
        self.input_args = args
        self.input_kwargs = kwargs
        
    def set_output_for_any_input_in_last_stubbed_method(self, args):
        self.stubs_return_values[self.last_stubbed_method] = args
            
    def set_output_for_last_stubbed_method(self, args):
        self.stubs_return_values[self.args_matcher.method_key(
             self.last_stubbed_method, self.input_args, 
             self.input_kwargs)] = args
                
                
                
class _EmptyObject_():
    """
    Internal framework class. Intended to be used as the original_object
    for empty_spy and empty_mock
    """
    asserting_on_method = None
    _empty_object__ = True
    def method(self, *args, **kwargs):
        pass
    
    def __getattr__(self, attr):
        self.method.im_func.func_name = attr
        return self.method
                
