from core import _CallsRegistry_, _StubsRepository_
from core import _EmptyObject_, _Introspector_, _MethodHandler_
from core import UnexpectedBehavior, WrongApiUsage, ApiMismatch
from core import ArgsDontMatch, ANY_ARG
 
"""
Authors: www.iExpertos.com
   - Carlos Ble (www.carlosble.com)

License: Apache 2 (http://www.apache.org/licenses/LICENSE-2.0.html)

Project home: http://www.pydoubles.org
"""
                
                
class ProxySpy(object):
    """
    This test double is just an _interceptor to the original object.
    It watches the calls to the original, recording what happens
    so that we can make assertions on the calls that were made.
    The actual methods in the original object are executed, they
    are not mocked or stubbed by default.
    """
    def __init__(self, original_instance):
        self.registry = _CallsRegistry_()
        self.introspector = _Introspector_()
        self.stubs = _StubsRepository_()
        self.stubbed_method_names = []
        self.original_instance = original_instance
        self.original_instance._double = self
        self.asserting_on_method = None
        self.invoked_method_name = None

    def __getattr__(self, attr):
        return _MethodHandler_(self, attr)
        
    def _before_method(self, args, kwargs):
        self._replace_possible_alias_method_name()
        self.registry.register_call(self.invoked_method_name, 
                                            *args, **kwargs)
        
    def __remove_interceptor_from_call_args(self, args):
        return args[1:]
    
    def _on_method_call(self, *args, **kwargs):
        args = self.__remove_interceptor_from_call_args(args)
        self._before_method(args, kwargs)
        if self.stubs.is_matching_stubbed_method(self.invoked_method_name, 
                                    args, kwargs):
           return self.stubs.return_value_given_input(
                         self.invoked_method_name, args, kwargs)
        return self._invoke_method(args, kwargs)
        
    def _invoke_method(self, args, kwargs):
        original_method = getattr(self.original_instance, 
                                  self.invoked_method_name)
        return original_method(*args, **kwargs)
        
    def was_called(self, method):
        return self.registry.was_registered(
                    self.introspector.method_name(method))

    def _is_asserting_that_call_was_made(self):
        return self.asserting_on_method is not None
    
    def with_args(self, *args, **kwargs):
        if self._is_asserting_that_call_was_made():
           self.registry.match_call(self.introspector.method_name(
                                    self.asserting_on_method),
                                    *args, **kwargs)
        else:
           self.stubs.set_input_for_last_stubbed_method(args, kwargs)
        return self
        
    def stub_out(self, method):
        method_name = self.introspector.method_name(method)
        self.stubs.create_stub(method_name)
        self.stubbed_method_names.append(method_name)
        
    def _set_output_for_stub(self, args):
        if self.stubs.will_stub_any_input():
           self.stubs.set_output_for_any_input_in_last_stubbed_method(
                                                                args)
        else:
           self.stubs.set_output_for_last_stubbed_method(args)
           
    def then_return(self, args):
        self._set_output_for_stub(args)

    def then_raise(self, exception_instance):
        method = method_raising(exception_instance)
        self._set_output_for_stub(method)
        
    def _replace_possible_alias_method_name(self):
        for method_name in self.stubbed_method_names:
            if self.introspector.are_synonymous(self.original_instance, 
               self.invoked_method_name, method_name):
                   self.invoked_method_name = method_name
                   return
        

class Spy(ProxySpy):
    """
    Works like a ProxySpy but methods are stubbed returning
    nothing by default, rather than calling the doubled object.
    The spy will check if the API consumed by the user matches
    the actual API in the doubled object.
    """
    DEFAULT_RETURN_VALUE_FOR_STUBBED_METHOD = None
    
    def _api_mismatch_msg(self, original_args, args):
        return "The number of arguments in the actual object <%s> don't match the actual call on the double <%s>" % (
               str(tuple(original_args)).replace("'self', ", ""),
               str(args))
            
    def _check_api_match(self, args, kwargs):
        original_method = getattr(self.original_instance, 
                                  self.invoked_method_name)
        args_plus_kwargs_spec_len, kwargs_spec_len, arg_spec = \
          self.introspector.original_method_signature(original_method) 
        if len(args) + len(kwargs) != args_plus_kwargs_spec_len and \
           len(args) != args_plus_kwargs_spec_len - kwargs_spec_len:
             msg = self._api_mismatch_msg(arg_spec.args, args) 
             raise ApiMismatch(msg)
        
    def _before_method(self, args, kwargs):
        super(Spy, self)._before_method(args, kwargs)
        if not hasattr(self.original_instance, "_empty_object__"):
           self._check_api_match(args, kwargs)        
        
    def _invoke_method(self, args, kwargs):
        return self.DEFAULT_RETURN_VALUE_FOR_STUBBED_METHOD

    
    
class Mock(Spy):
    """
    Any method that is going to be invoked on this object must
    be expected through the expect_call method before invoking it. 
    Otherwise, an exception will be thrown. 
    """
    def __init__(self, obj_instance):
        super(Mock, self).__init__(obj_instance)
        self.satisfied_expectations = []
        self.expectations_with_defined_arguments = []
        
    def add_expectation(self, method):
        self.stub_out(method)
        
    def with_args(self, *args, **kwargs):
        super(Mock, self).with_args(*args, **kwargs)
        self.expectations_with_defined_arguments.append(self.stubbed_method_names[-1])
        self.stubs.set_output_for_last_stubbed_method(None)
        return self
        
    def _is_expectation_without_arguments_defined(self):
        return self.invoked_method_name in self.stubbed_method_names and \
               self.invoked_method_name not in self.expectations_with_defined_arguments
          
    def _invoked_method_was_expected(self):
        return self.invoked_method_name in self.stubbed_method_names
                       
    def __matches_expectation(self, args, kwargs):
        return self._invoked_method_was_expected() and \
               (self._is_expectation_without_arguments_defined() or \
                 self.stubs.is_stub_for_this_input(
                  self.invoked_method_name, args, kwargs))

    def _before_method(self, args, kwargs):
        super(Mock, self)._before_method(args, kwargs)
        if self.__matches_expectation(args, kwargs):
           self.satisfied_expectations.append(self.invoked_method_name)
        else:
           raise UnexpectedBehavior(
                 "This call wasn't expected: '%s' . Expected methods are: %s" % (
                 self.invoked_method_name, 
                 str(self.stubbed_method_names)))
                          
    def returning(self, args):
        return super(Mock, self).then_return(args)
    
    def assert_that_is_satisfied(self):
        if len(self.stubbed_method_names) > len(self.satisfied_expectations):
           raise UnexpectedBehavior("Defined expectations were not satisfied. Registered calls are:" + self.registry.show_registered_calls())

       
# PUBLIC STATIC METHODS:        

def proxy_spy(obj_instance):
    """
    Creates a spy objet that records the calls made but passes them
    to the actual object.
    """
    return ProxySpy(obj_instance)
        
def spy(obj_instance):
    """
    Creates a spy object based on the given object instance
    """
    return Spy(obj_instance)

def empty_stub():
    """
    Creates a stub object in which you can dynamically add any method  
    """
    return empty_spy()

def stub(obj_instance):
    """
    Creates a stub object based on the given object instance
    """
    return spy(obj_instance)

def empty_spy():
    """
    Creates a spy object but will not check any API match
    """
    return Spy(_EmptyObject_())

def mock(obj_instance):
    """
    Creates a mock objet based on the given object instance
    """
    return Mock(obj_instance)

def empty_mock():
    """
    Creates a mock objet but will not check any API match
    """
    return mock(_EmptyObject_())

def method_returning(return_value):
    """
    Creates a method stub, able to receive anything and return 
    the given return_value
    """
    return lambda *args, **kwargs: return_value

def method_raising(exception_instance):
    """
    Creates a method stub, which raises the given exception
    """
    def raise_on_call(*args, **kwargs):
        raise exception_instance
    return raise_on_call    
    
def expect_call(method):
    """
    Define behavior in a mock object
    """
    double = _Introspector_().double_instance_from_method(method)
    double.add_expectation(method)
    return double
    
def when(method):
    """
    Define behavior in a stub or spy object
    """
    double = _Introspector_().double_instance_from_method(method)
    double.stub_out(method)
    return double  
        
def assert_that_was_called(method):
    """
    Verify the behavior in a spy. Use this with spies only.
    For mock objects use: mock_instance.assert_that_is_satisfied() 
    """
    try:
        double = _Introspector_().double_instance_from_method(method) 
        if double.was_called(method):
           double.asserting_on_method = method
           return double
        raise UnexpectedBehavior("The method was not called")
    except AttributeError, e:
        raise WrongApiUsage(
            "Make sure you call assert, passing in a method from a test double: (double.method)")
        
