# -*- coding: utf-8 -*-

import unittest
from pyDoubles.framework import *
from pyDoubles.core import _Introspector_, _Combinator_ 
from pyDoubles.core import _ArgsMatcher_, _CallsRegistry_
from pyDoubles.core import _ArgsAdapter_

"""
Author: Carlos Ble (www.carlosble.com)
License: Apache 2 (http://www.apache.org/licenses/LICENSE-2.0.html)
Project home: https://bitbucket.org/carlosble/pydoubles
"""

class SomeException(Exception):
    pass
        
class Collaborator():
    """
    The original object we double in tests
    """
    def hello(self):
        return "hello"
    
    def something(self):
        return "ok"
    
    def one_arg_method(self, arg1):
        return arg1
    
    def two_args_method(self, arg1, arg2):
        return arg1 + arg2
    
    def kwarg_method(self, key_param=False):
        return key_param
    
    def mixed_method(self, arg1, key_param=False):
        return key_param + arg1

    def void_method(self):
        pass
    
    alias_method = one_arg_method


class ProxySpyTests(unittest.TestCase):
    
    def setUp(self):
        self.spy = proxy_spy(Collaborator())
            
    def test_assert_was_called(self):
        self.spy.hello()
        
        assert_that_was_called(self.spy.hello)

    def test_assert_was_called_on_any_method(self):
        self.spy.something()
        
        assert_that_was_called(self.spy.something)

    def test_assert_needs_always_a_method_from_a_double(self):
        self.failUnlessRaises(WrongApiUsage,
                assert_that_was_called, self.spy)

    def test_assert_needs_always_a_method_from_a_double_not_the_original(self):
        self.failUnlessRaises(WrongApiUsage,
             assert_that_was_called, Collaborator().hello)

    def test_one_method_called_other_wasnt(self):
        self.spy.something()
        
        self.failUnlessRaises(UnexpectedBehavior, 
             assert_that_was_called, self.spy.hello)                

    def test_two_methods_called_assert_on_the_first(self):
        self.spy.hello()
        self.spy.something()
        
        assert_that_was_called(self.spy.hello)

    def test_get_method_name(self):
        name = _Introspector_().method_name(self.spy.hello)
        
        self.assertEquals("hello", name)
                
    def test_call_original_method(self):
        self.assertEquals("ok", self.spy.something())

    def test_get_instance_from_method(self):
        spy_found = _Introspector_().double_instance_from_method(self.spy.hello)
        
        self.assertEquals(self.spy, spy_found)
                
    def test_assert_was_called_when_wasnt(self):
        self.failUnlessRaises(UnexpectedBehavior, 
             assert_that_was_called, self.spy.hello)                

    def test_was_called_with_same_parameters(self):
        self.spy.one_arg_method(1)
        
        assert_that_was_called(self.spy.one_arg_method).with_args(1)

    def test_was_called_with_same_parameters_in_variables(self):
        arg1 = 1
        self.spy.one_arg_method(arg1)
        
        assert_that_was_called(self.spy.one_arg_method).with_args(1)
    
    def test_was_called_with_same_parameters_when_not(self):
        self.spy.one_arg_method(1)
        args_checker = assert_that_was_called(self.spy.one_arg_method)
        
        self.failUnlessRaises(ArgsDontMatch,
            args_checker.with_args, 2)
        
    def test_get_args_of_last_call(self):
        self.spy.one_arg_method(1)
        
        self.spy.registry.match_call("one_arg_method", 1)
    
    def test_was_called_with_same_params_but_no_params_accepted(self):
        self.spy.hello()
        args_checker = assert_that_was_called(self.spy.hello)
        
        self.failUnlessRaises(ArgsDontMatch,
            args_checker.with_args, "something")

    def test_was_called_with_several_parameters(self):
        self.spy.two_args_method(1, 2)
        args_checker = assert_that_was_called(self.spy.two_args_method)
       
        args_checker.with_args(1, 2) 

    def test_was_called_with_parameters_not_matching(self):
        self.spy.one_arg_method(1)
        args_checker = assert_that_was_called(self.spy.one_arg_method)

        self.failUnlessRaises(ArgsDontMatch,
            args_checker.with_args, "2") 

    def test_was_called_with_keyed_args_not_matching(self):
        self.spy.kwarg_method(key_param="foo")
        args_checker = assert_that_was_called(self.spy.kwarg_method)

        self.failUnlessRaises(ArgsDontMatch,
            args_checker.with_args, key_param="bar")

    def test_was_called_with_keyed_args_matching(self):
        self.spy.kwarg_method(key_param="foo")
        assert_that_was_called(self.spy.kwarg_method).with_args(
                                    key_param="foo")
        
    def test_recorded_call_params_are_displayed(self):
        self.spy.kwarg_method(key_param="foo")
        try:
            assert_that_was_called(self.spy.kwarg_method
                                   ).with_args("bar")
        except ArgsDontMatch, e:
            self.assertTrue(str(e).find("foo") != -1, str(e))

    def test_stub_out_method(self):
        when(self.spy.one_arg_method).then_return(3)
        
        self.assertEquals(3, self.spy.one_arg_method(5))
        
        assert_that_was_called(self.spy.one_arg_method).with_args(5)

    def test_stub_out_method_returning_a_list(self):
        when(self.spy.one_arg_method).then_return([1,2,3])
        
        self.assertEquals([1,2,3], self.spy.one_arg_method(5))
        
        assert_that_was_called(self.spy.one_arg_method).with_args(5)

    def test_stub_out_method_with_args(self):
        when(self.spy.one_arg_method).with_args(2).then_return(3)
        
        self.assertEquals(3, self.spy.one_arg_method(2))
        
        assert_that_was_called(self.spy.one_arg_method).with_args(2)

    def test_stub_out_method_with_args_calls_actual(self):
        when(self.spy.one_arg_method).with_args(2).then_return(3)
        
        self.assertEquals(4, self.spy.one_arg_method(4))
        
        assert_that_was_called(self.spy.one_arg_method).with_args(4)

    def test_stub_out_method_with_several_inputs(self):
        when(self.spy.one_arg_method).with_args(2).then_return(3)
        when(self.spy.one_arg_method).with_args(3).then_return(4)
        
        self.assertEquals(3, self.spy.one_arg_method(2))
        self.assertEquals(4, self.spy.one_arg_method(3))

    def test_recorded_calls_work_on_several_stubs(self):
        when(self.spy.one_arg_method).with_args(2).then_return(3)
        when(self.spy.one_arg_method).with_args(3).then_return(4)
        
        self.spy.one_arg_method(2)        
        self.spy.one_arg_method(3)
        assert_that_was_called(self.spy.one_arg_method).with_args(2)
        assert_that_was_called(self.spy.one_arg_method).with_args(3)

    def test_stub_with_ignore_arguments_overrides_specific_arguments(self):
        when(self.spy.one_arg_method).with_args(2).then_return(3)
        when(self.spy.one_arg_method).then_return(1000)
        
        self.assertEquals(1000, self.spy.one_arg_method(2))
        self.assertEquals(1000, self.spy.one_arg_method(8))

    def test_stub_with_kwargs(self):
        when(self.spy.kwarg_method).with_args(key_param=2
                                            ).then_return(3)
        
        self.assertEquals(3, self.spy.kwarg_method(key_param=2))
        self.assertEquals(6, self.spy.kwarg_method(key_param=6))
        
    def test_stub_raising_exception(self):
        when(self.spy.hello).then_raise(SomeException())
        try:
            self.spy.hello()
            self.fail("not raised")
        except SomeException:
            pass            
        
    def test_be_able_to_return_objects(self):
        when(self.spy.one_arg_method).then_return(Collaborator())
        
        collaborator = self.spy.one_arg_method(1)
        
        self.assertEquals(1, collaborator.one_arg_method(1))
        
    def test_any_arg_matcher(self):
        when(self.spy.two_args_method).with_args(1, ANY_ARG).then_return(1000)
        
        self.assertEquals(1000, self.spy.two_args_method(1,2))                                
        self.assertEquals(1000, self.spy.two_args_method(1,5))

# TODO: implement this:
#    def test_any_arg_matcher_with_kwargs(self):
#        when(self.spy.kwarg_method).with_args(key_param=ANY_ARG).then_return(1000)
#        
#        self.assertEquals(1000, self.spy.kwarg_method(key_param=2))                                

    def test_any_arg_matcher_was_called(self):
        when(self.spy.two_args_method).with_args(1, 2).then_return(1000)
        
        self.spy.two_args_method(1,2)                                
        
        assert_that_was_called(self.spy.two_args_method
                               ).with_args(1, ANY_ARG)                

    def test_stub_works_with_alias_method(self):
        when(self.spy.one_arg_method).with_args(1).then_return(1000)
        
        self.spy.alias_method(1)
        assert_that_was_called(self.spy.one_arg_method
                               ).with_args(1)                

class CombinatorTests(unittest.TestCase):
    def setUp(self):
        self.combinator = _Combinator_()
    
    def test_replace_any_in_empty_list(self):
        self.assertEquals(((),), 
                self.combinator.all_args_combinations([]))
    
    def test_replace_any_in_simple_list(self):
        tuples = self.combinator.all_args_combinations((1,))
        self.assertTrue((ANY_ARG,) in tuples, str(tuples))
        self.assertTrue((1,) in tuples, str(tuples)) 
                
    def test_replace_any_in_large_list(self):
        tuples = self.combinator.all_args_combinations((1,2))
        self.assertTrue((1, ANY_ARG) in tuples, str(tuples))
        self.assertTrue((ANY_ARG, 2) in tuples, str(tuples))
        self.assertTrue((1,2) in tuples, str(tuples)) 
        
    def test_more_than_one_param_is_any(self):
        tuples = self.combinator.all_args_combinations((1,2))
        self.assertTrue((ANY_ARG, ANY_ARG) in tuples, str(tuples))

    def test_only_one_param_is_not_any(self):
        tuples = self.combinator.all_args_combinations((1,2,3))
        self.assertTrue((ANY_ARG, 2, ANY_ARG) in tuples, str(tuples))
        self.assertTrue((ANY_ARG, ANY_ARG, 3) in tuples, str(tuples))
        self.assertTrue((1, ANY_ARG, ANY_ARG) in tuples, str(tuples))
    
    
class ArgsMatcherTests(unittest.TestCase):
    
    def setUp(self):
        self.matcher = _ArgsMatcher_()
                
    def test_is_method_match(self):
        name = "one_method"
        dict = {self.matcher.method_key(name, (1, ANY_ARG), {}): 1}
        self.assertTrue(self.matcher.is_method_match(
                        dict, name, (1,1), {}))    

    def test_is_method_match_with_kwargs(self):
        name = "one_method"
        dict = {self.matcher.method_key(name, (), {'kw':1}): 1}
        self.assertTrue(self.matcher.is_method_match(
                        dict, name, (), {'kw':1}))
        
    def test_call_args_match(self):
        self.assertTrue(
            self.matcher.do_call_args_match(
                (1,2), ['args=(1, 2)', 'kwargs={}']))    

    def test_call_args_match_with_any(self):
        self.assertTrue(
            self.matcher.do_call_args_match(
                (1, ANY_ARG), ['args=(1, 2)', 'kwargs={}']))    

    def test_call_args_match_with_spaces(self):
        self.assertTrue(
            self.matcher.do_call_args_match(
                ("sup dude"   , ANY_ARG), ['args=("sup dude" ,2)', 'kwargs={}']))    

    def test_call_args_match_with_apostrofes(self):
        self.assertTrue(
            self.matcher.do_call_args_match(
                ("don't", ANY_ARG), ['args=("don\'t", 2)', 'kwargs={}']))    


class ArgsAdapterTests(unittest.TestCase):
    def setUp(self):
        self.adapter = _ArgsAdapter_()
        
    def test_call_args_match_with_unicode(self):
        self.assertEquals(['OK', '2'],
            self.adapter.transform((u"OK", 2)))            

    def test_get_call_args(self):
        self.assertEquals(['(1, 2)'], 
            self.adapter.clean_call_args(['args=(1, 2)', 'kwargs={}']))

    def test_get_call_args_when_args_contain_equals(self):
        self.assertEquals(["(u'/action/==MQQ61', 2)"], 
            self.adapter.clean_call_args(["args=(u'/action/==MQQ61', 2)", 'kwargs={}']))


class SpyTests(unittest.TestCase):
    def setUp(self):
        self.spy = spy(Collaborator())
        
    def test_override_original_method(self):
        self.assertTrue(self.spy.hello() is None)
        
        assert_that_was_called(self.spy.hello)        
        
    def test_spy_can_work_from_empty_object(self):
        self.spy = empty_spy()
        self.assertTrue(self.spy.hello() is None)
        
        assert_that_was_called(self.spy.hello)
        
    def test_spy_based_on_object_must_check_api_match(self):
        try:
            self.spy.hello("unexpected argument")
            self.fail('Expection should raise: Actual objet does not accept parameters')
        except ApiMismatch:
            pass

    def test_check_api_match_with_kwargs(self):
        self.assertTrue(self.spy.mixed_method(1, key_param=2) is None)

    def test_check_api_match_with_kwargs_not_used(self):
        self.assertTrue(self.spy.mixed_method(1) is None)
        
    def test_check_api_match_with_kwargs_not_matching(self):
        try:
            self.spy.mixed_method(1,2,3)
            self.fail('Api mismatch not detected')
        except ApiMismatch:
            pass

    def test_match_call_with_unicode_and_non_ascii_chars(self):
        non_ascii  = u'España'
        self.spy.one_arg_method(non_ascii)
        
        assert_that_was_called(self.spy.one_arg_method).with_args(
                                                        non_ascii)    

    def test_stub_methods_can_be_handled_separately(self):
        when(self.spy.one_arg_method).with_args(1).then_return(1000)
        when(self.spy.two_args_method).with_args(5,5).then_return(2000)
        handle1 = self.spy.one_arg_method
        handle2 = self.spy.two_args_method
        self.assertEquals(1000, handle1(1))
        self.assertEquals(2000, handle2(5,5))
        
        assert_that_was_called(handle1).with_args(1)
        assert_that_was_called(handle2).with_args(5,5)
                
        
class MockTests(unittest.TestCase):
    
    def setUp(self):
        self.mock = mock(Collaborator())
        
    def test_fail_on_unexpected_call(self):
        try:
            self.mock.hello()
            self.fail('UnexpectedBehavior should be raised')
        except UnexpectedBehavior:
            pass
        
    def test_define_expectation_and_call_method(self):
        expect_call(self.mock.hello)
        self.assertTrue(self.mock.hello() is None)
        
    def test_define_several_expectatiosn(self):
        expect_call(self.mock.hello)
        expect_call(self.mock.one_arg_method)
        
        self.assertTrue(self.mock.hello() is None)
        self.assertTrue(self.mock.one_arg_method(1) is None)
        
    def test_define_expectation_args(self):
        expect_call(self.mock.one_arg_method).with_args(1)
        self.assertTrue(self.mock.one_arg_method(1) is None)
        
    def test_define_expectation_args_and_fail(self):
        expect_call(self.mock.one_arg_method).with_args(1)
        try:
            self.mock.one_arg_method(2)
            self.fail('Unexpected call')
        except UnexpectedBehavior:
            pass
        
    def test_several_expectations_with_args(self):
        expect_call(self.mock.one_arg_method).with_args(1)
        expect_call(self.mock.two_args_method).with_args(2,3)
        
        self.assertTrue(self.mock.one_arg_method(1) is None)
        self.assertTrue(self.mock.two_args_method(2,3) is None)
    
    def test_expect_call_returning_value(self):
        expect_call(self.mock.one_arg_method).with_args(1).returning(1000)
        
        self.assertEquals(1000, self.mock.one_arg_method(1))    
        
    def test_assert_expectations_are_satisfied(self):
        expect_call(self.mock.hello)
        try:
            self.mock.assert_that_is_satisfied()
            self.fail('Not satisfied!')
        except UnexpectedBehavior:
            pass
        
    def test_assert_satisfied_when_it_really_is(self):
        expect_call(self.mock.hello)
        self.mock.hello()
        self.mock.assert_that_is_satisfied()
        
    def test_number_of_calls_doesnt_matter(self):
        expect_call(self.mock.hello)
        self.mock.hello()
        self.mock.hello()
        self.mock.assert_that_is_satisfied()
        
    def test_using_when_or_expect_call_without_double(self):
        self.failUnlessRaises(WrongApiUsage, 
                        expect_call, Collaborator())
                
    def test_expectations_on_synonyms(self):
        expect_call(self.mock.one_arg_method)
        
        self.mock.alias_method(1)
        
        self.mock.assert_that_is_satisfied()
        
        
        
class MockFromEmptyObjectTests(unittest.TestCase):
    def setUp(self):
        self.mock = empty_mock()
                
    def test_mock_can_work_from_empty_object(self):
        expect_call(self.mock.hello)
        
        self.mock.hello()
        
        self.mock.assert_that_is_satisfied()
        
    def test_several_expectations_in_empty_mock(self):
        expect_call(self.mock.hello)
        expect_call(self.mock.one_arg_method).with_args(1)
        
        self.mock.hello()
        self.mock.one_arg_method(1)
        
        self.mock.assert_that_is_satisfied()
        
    def test_several_expectations_with_args_in_empty_mock(self):
        expect_call(self.mock.one_arg_method).with_args(1)
        expect_call(self.mock.one_arg_method).with_args(2)
        
        self.assertTrue(self.mock.one_arg_method(1) is None)
        self.assertTrue(self.mock.one_arg_method(2) is None)
        
        self.mock.assert_that_is_satisfied()
        
        
class StubMethodsTests(unittest.TestCase):
    
    def setUp(self):
        self.collaborator = Collaborator()
        
    def test_method_returning_value(self):
        self.collaborator.hello = method_returning("bye")
        
        self.assertEquals("bye", self.collaborator.hello())
        
    def test_method_args_returning_value(self):
        self.collaborator.one_arg_method = method_returning("bye")
        
        self.assertEquals("bye", self.collaborator.one_arg_method(1))
        
    def test_method_raising_exception(self):
        self.collaborator.hello = method_raising(SomeException())
        try:
            self.collaborator.hello()
            self.fail("exception not raised")
        except SomeException:
            pass
        
class CallsRegistryTests(unittest.TestCase):
    
    def test_match_call_once_its_been_registered(self):
        registry = _CallsRegistry_()
        registry.register_call("methodX", 1,2,3)
        registry.match_call("methodX", 1,2,3)
        
            
        
        

if __name__ == "__main__":
    print "Use nosetest to run this tests: nosetest unit.py"





















